@@ -2,9 +2,9 @@
    Copyright (C) 1997-9 Eddie Kohler, eddietwo@lcs.mit.edu
    This file is part of the GIF library.
 
-   The GIF library is free software*. It is distributed under the GNU Public
-   License, version 2 or later; you can copy, distribute, or alter it at will,
-   as long as this notice is kept intact and this source code is made
+   The GIF library is free software*. It is distributed under the GNU General
+   Public License, version 2 or later; you can copy, distribute, or alter it
+   at will, as long as this notice is kept intact and this source code is made
    available. There is no warranty, express or implied.
 
    *The LZW compression method used by GIFs is patented. Unisys, the patent
@@ -50,12 +50,14 @@ typedef struct {
 typedef struct Gif_Reader {
   
   FILE *f;
-  byte *v;
+  const byte *v;
   u_int32_t w;
+  u_int32_t length;
   int is_record;
   int is_eoi;
   byte (*byte_getter)(struct Gif_Reader *);
   void (*block_getter)(byte *, u_int32_t, struct Gif_Reader *);
+  u_int32_t (*offseter)(struct Gif_Reader *);
   int (*eofer)(struct Gif_Reader *);
   
 } Gif_Reader;
@@ -64,6 +66,7 @@ typedef struct Gif_Reader {
 #define gifgetc(grr)	((char)(*grr->byte_getter)(grr))
 #define gifgetbyte(grr) ((*grr->byte_getter)(grr))
 #define gifgetblock(ptr, size, grr) ((*grr->block_getter)(ptr, size, grr))
+#define gifgetoffset(grr) ((*grr->offseter)(grr))
 #define gifeof(grr)	((*grr->eofer)(grr))
 
 static inline u_int16_t
@@ -85,7 +88,13 @@ file_byte_getter(Gif_Reader *grr)
 static void
 file_block_getter(byte *p, u_int32_t s, Gif_Reader *grr)
 {
-  fread(p, s, 1, grr->f);
+  fread(p, 1, s, grr->f);
+}
+
+static u_int32_t
+file_offseter(Gif_Reader *grr)
+{
+  return ftell(grr->f);
 }
 
 static int
@@ -109,6 +118,12 @@ record_block_getter(byte *p, u_int32_t s, Gif_Reader *grr)
   grr->w -= s, grr->v += s;
 }
 
+static u_int32_t
+record_offseter(Gif_Reader *grr)
+{
+  return grr->length - grr->w;
+}
+
 static int
 record_eofer(Gif_Reader *grr)
 {
@@ -117,13 +132,15 @@ record_eofer(Gif_Reader *grr)
 
 
 static void
-make_data_reader(Gif_Reader *grr, byte *data, u_int32_t length)
+make_data_reader(Gif_Reader *grr, const byte *data, u_int32_t length)
 {
   grr->v = data;
+  grr->length = length;
   grr->w = length;
   grr->is_record = 1;
   grr->byte_getter = record_byte_getter;
   grr->block_getter = record_block_getter;
+  grr->offseter = record_offseter;
   grr->eofer = record_eofer;
 }
 
@@ -189,7 +206,7 @@ read_image_block(Gif_Reader *grr, byte *buffer, int *bit_pos_store,
       bit_length -= i * 8;
     }
     block_len = gifgetbyte(grr);
-    GIF_DEBUG(("image_block(%d)", block_len));
+    GIF_DEBUG(("\nimage_block(%d)", block_len));
     if (block_len == 0) return 0;
     gifgetblock(buffer + bit_length / 8, block_len, grr);
     bit_length += block_len * 8;
@@ -227,7 +244,7 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
   gfc->decodepos = 0;
   
   min_code_size = gifgetbyte(grr);
-  GIF_DEBUG(("min_code_size(%d)", min_code_size));
+  GIF_DEBUG(("\n\nmin_code_size(%d)", min_code_size));
   if (min_code_size >= GIF_MAX_CODE_BITS) {
     gif_read_error(gfc, "min_code_size too big");
     min_code_size = GIF_MAX_CODE_BITS - 1;
@@ -278,7 +295,7 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
     code = (Gif_Code)((accum >> (bit_position % 8)) & CUR_CODE_MASK);
     bit_position += bits_needed;
     
-    GIF_DEBUG(("code(%d)", code));
+    GIF_DEBUG(("%d", code));
     
     /* CHECK FOR SPECIAL OR BAD CODES: clear_code, eoi_code, or a code that is
      * too large. */
@@ -336,11 +353,11 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
   
   /* read blocks until zero-length reached. */
   i = gifgetbyte(grr);
-  GIF_DEBUG(("after_image(%d)", i));
+  GIF_DEBUG(("\nafter_image(%d)\n", i));
   while (i > 0) {
     gifgetblock(buffer, i, grr);
     i = gifgetbyte(grr);
-    GIF_DEBUG(("after_image(%d)", i));
+    GIF_DEBUG(("\nafter_image(%d)\n", i));
   }
   
   /* zero-length block reached. */
@@ -403,7 +420,7 @@ static int
 read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
 {
   if (grr->is_record) {
-    byte *first = grr->v;
+    const byte *first = grr->v;
     u_int32_t pos;
     
     /* scan over image */
@@ -417,7 +434,7 @@ read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
     
     gfi->compressed_len = pos;
     if (read_flags & GIF_READ_CONST_RECORD) {
-      gfi->compressed = first;
+      gfi->compressed = (byte *)first;
       gfi->free_compressed = 0;
     } else {
       gfi->compressed = Gif_NewArray(byte, gfi->compressed_len);
@@ -724,7 +741,7 @@ read_gif(Gif_Reader *grr, int read_flags,
   Gif_Image *new_gfi;
   Gif_Context gfc;
   int extension_position = 0;
-  int ok = 0;
+  int unknown_block_type = 0;
   
   if (gifgetc(grr) != 'G' ||
       gifgetc(grr) != 'I' ||
@@ -759,7 +776,7 @@ read_gif(Gif_Reader *grr, int read_flags,
     switch (block) {
       
      case ',': /* image block */
-      GIF_DEBUG(("imageread"));
+      GIF_DEBUG(("imageread %d", gfs->nimages));
       
       gfi->identifier = last_name;
       last_name = 0;
@@ -772,18 +789,12 @@ read_gif(Gif_Reader *grr, int read_flags,
       new_gfi = Gif_NewImage();
       if (!new_gfi) goto done;
       
-      if (read_flags & GIF_READ_NETSCAPE_WORKAROUND) {
-	new_gfi->transparent = gfi->transparent;
-	new_gfi->delay = gfi->delay;
-	new_gfi->disposal = gfi->disposal;
-      }
       gfi = new_gfi;
       extension_position++;
       break;
       
      case ';': /* terminator */
-      GIF_DEBUG(("term"));
-      ok = 1;
+      GIF_DEBUG(("term\n"));
       goto done;
       
      case '!': /* extension */
@@ -815,8 +826,13 @@ read_gif(Gif_Reader *grr, int read_flags,
       break;
       
      default:
-      gif_read_error(&gfc, "unknown block type");
-      break;
+       if (!unknown_block_type) {
+	 char buf[256];
+	 sprintf(buf, "unknown block type %d at file offset %d", block, gifgetoffset(grr) - 1);
+	 gif_read_error(&gfc, buf);
+	 unknown_block_type = 1;
+       }
+       break;
       
     }
     
@@ -833,12 +849,7 @@ read_gif(Gif_Reader *grr, int read_flags,
   Gif_DeleteArray(gfc.prefix);
   Gif_DeleteArray(gfc.suffix);
   Gif_DeleteArray(gfc.length);
-  if (ok)
-    return gfs;
-  else {
-    Gif_DeleteStream(gfs);
-    return 0;
-  }
+  return gfs;
 }
 
 
@@ -852,12 +863,13 @@ Gif_FullReadFile(FILE *f, int read_flags,
   grr.is_record = 0;
   grr.byte_getter = file_byte_getter;
   grr.block_getter = file_block_getter;
+  grr.offseter = file_offseter;
   grr.eofer = file_eofer;
   return read_gif(&grr, read_flags, h, hthunk);
 }
 
 Gif_Stream *
-Gif_FullReadRecord(Gif_Record *gifrec, int read_flags,
+Gif_FullReadRecord(const Gif_Record *gifrec, int read_flags,
 		   Gif_ReadErrorHandler h, void *hthunk)
 {
   Gif_Reader grr;
@@ -869,6 +881,22 @@ Gif_FullReadRecord(Gif_Record *gifrec, int read_flags,
 }
 
 
+#undef Gif_ReadFile
+#undef Gif_ReadRecord
+
+Gif_Stream *
+Gif_ReadFile(FILE *f)
+{
+  return Gif_FullReadFile(f, GIF_READ_UNCOMPRESSED, 0, 0);
+}
+
+Gif_Stream *
+Gif_ReadRecord(const Gif_Record *gifrec)
+{
+  return Gif_FullReadRecord(gifrec, GIF_READ_UNCOMPRESSED, 0, 0);
+}
+
+
 #ifdef __cplusplus
 }
 #endif
