@@ -2,10 +2,9 @@
 #include "xwrits.h"
 #include <stdlib.h>
 #include <stdio.h>
+#include <assert.h>
 
-
-Alarm *alarms;
-
+static Alarm alarm_sentinel;
 
 /*****************************************************************************/
 /*  Idle functions							     */
@@ -25,30 +24,31 @@ static unsigned long key_press_selected_count;
 int
 x_error_handler(Display *d, XErrorEvent *error)
 {
-  if (d != display ||
-      (error->error_code != BadWindow && error->error_code != BadDrawable)) {
+  if (error->error_code != BadWindow && error->error_code != BadDrawable) {
     char buffer[BUFSIZ];
-    XGetErrorText(display, error->error_code, buffer, BUFSIZ);
+    XGetErrorText(d, error->error_code, buffer, BUFSIZ);
     fprintf(stderr, "X Error of failed request: %s\n", buffer);
     abort();
   }
   
   /* Maybe someone created a window then destroyed it immediately!
      I don't think there's any way of working around this. */
-  unschedule_data(A_IDLE_SELECT, (void *)((Window)error->resourceid));
+  unschedule_data(A_IDLE_SELECT, (void *)error->display,
+		  (void *)((Window)error->resourceid));
   return 0;
 }
 
 
 void
-watch_keystrokes(Window w, struct timeval *now)
+watch_keystrokes(Display *display, Window w, const struct timeval *now)
 {
   Window root, parent, *children;
   unsigned i, nchildren;
   Alarm *a;
   
-  if (window_to_hand(w, 1))
-    return; /* don't need to worry about our own windows */
+  /* Don't pay attention to our own windows */
+  if (window_to_hand(find_port(display), w, 1))
+    return;
   
   if (XQueryTree(display, w, &root, &parent, &children, &nchildren) == 0)
     return; /* the window doesn't exist */
@@ -59,19 +59,18 @@ watch_keystrokes(Window w, struct timeval *now)
   /* This code ensures that at least register_keystrokes_delay elapses before
      we listen for KeyPress events on the window. We want to wait so we can
      make sure the window selects them first. */
-  a = new_alarm_data(A_IDLE_SELECT, (void *)w);
+  a = new_alarm_data(A_IDLE_SELECT, (void *)display, (void *)w);
   xwADDTIME(a->timer, *now, register_keystrokes_delay);
   schedule(a);
   
   for (i = 0; i < nchildren; i++)
-    watch_keystrokes(children[i], now);
+    watch_keystrokes(display, children[i], now);
   
   if (children) XFree(children);
 }
 
-
 void
-register_keystrokes(Window w)
+register_keystrokes(Display *display, Window w)
 {
   /* Before I only selected KeyPress if someone else had selected KeyPress on
      the window (indicated by the `or' of all_event_masks and
@@ -100,75 +99,78 @@ register_keystrokes(Window w)
 /*****************************************************************************/
 /*  Scheduling and alarm functions					     */
 
-
 Alarm *
-new_alarm_data(int action, void *data)
+new_alarm_data(int action, void *data1, void *data2)
 {
   Alarm *a = xwNEW(Alarm);
   a->action = action;
-  a->data = data;
+  a->data1 = data1;
+  a->data2 = data2;
+  a->scheduled = 0;
   return a;
 }
 
-
 Alarm *
-grab_alarm_data(int action, void *data)
+grab_alarm_data(int action, void *data1, void *data2)
 {
-  Alarm *prev = 0, *a;
-  for (a = alarms; a; prev = a, a = a->next)
-    if (a->action == action && a->data == data) {
-      if (prev)	prev->next = a->next;
-      else alarms = a->next;
+  Alarm *a;
+  for (a = alarm_sentinel.next; a != &alarm_sentinel; a = a->next)
+    if (a->action == action && a->data1 == data1 && a->data2 == data2) {
+      a->prev->next = a->next;
+      a->next->prev = a->prev;
+      a->scheduled = 0;
       return a;
     }
   return 0;
 }
 
-
 void
-destroy_alarm(Alarm *adestroy)
+destroy_alarm(Alarm *a)
 {
-  Alarm *a;
-  if (alarms == adestroy) alarms = adestroy->next;
-  for (a = alarms; a; a = a->next)
-    if (a->next == adestroy)
-      a->next = adestroy->next;
-  xfree(adestroy);
+  if (a->scheduled) {
+    a->prev->next = a->next;
+    a->next->prev = a->prev;
+  }
+  xfree(a);
 }
 
 
+void
+init_scheduler(void)
+{
+  alarm_sentinel.next = alarm_sentinel.prev = &alarm_sentinel;
+}
+
 void
 schedule(Alarm *newalarm)
 {
-  Alarm *prev = 0, *a = alarms;
-  while (a && xwTIMEGEQ(newalarm->timer, a->timer)) {
-    prev = a;
-    a = a->next;
-  }
-  if (prev) prev->next = newalarm;
-  else alarms = newalarm;
-  newalarm->next = a;
+  Alarm *a = alarm_sentinel.prev;
+  assert(!newalarm->scheduled);
+  while (a != &alarm_sentinel && xwTIMEGT(a->timer, newalarm->timer))
+    a = a->prev;
+  newalarm->prev = a;
+  newalarm->next = a->next;
+  a->next->prev = newalarm;
+  a->next = newalarm;
   newalarm->scheduled = 1;
 }
 
-
 void
-unschedule_data(int actions, void *data)
+unschedule_data(int actions, void *data1, void *data2)
 {
-  Alarm *prev = 0, *a = alarms;
-  while (a) {
+  Alarm *a = alarm_sentinel.next;
+  while (a != &alarm_sentinel) {
+    Alarm *n = a->next;
     
-    if ((a->action & actions) != 0 && (a->data == data || data == 0)) {
-      Alarm *n = a->next;
-      if (prev) prev->next = n;
-      else alarms = n;
+    if ((a->action & actions) != 0
+	&& (a->data1 == data1 || data1 == 0)
+	&& (a->data2 == data2 || data2 == 0)) {
+      a->prev->next = n;
+      n->prev = a->prev;
       xfree(a);
-      a = n;
-      continue;
     }
     
-    prev = a;
-    a = a->next;
+    a = n;
   }
 }
 
@@ -178,7 +180,7 @@ loopmaster(Alarmloopfunc alarm_looper, Xloopfunc x_looper)
 {
   struct timeval timeout, now, *timeoutptr;
   fd_set xfds;
-  int pending;
+  int pending, i;
   int ret_val = 0;
   
   /* 26 May 1998: Changed logic to avoid race conditions. Now we always flush
@@ -191,13 +193,18 @@ loopmaster(Alarmloopfunc alarm_looper, Xloopfunc x_looper)
   FD_ZERO(&xfds);
   
   while (1) {
-    while (alarms && xwTIMEGEQ(now, alarms->timer)) {
-      Alarm *a = alarms;
-      Hand *h = (Hand *)a->data;
+    while (1) {
+      Alarm *a = alarm_sentinel.next;
+      Hand *h;
       Gif_Stream *gfs;
-      
+
+      if (a == &alarm_sentinel || xwTIMEGT(a->timer, now))
+	break;
+
+      h = (Hand *)a->data1;
+      alarm_sentinel.next = a->next;
+      a->next->prev = &alarm_sentinel;
       a->scheduled = 0;
-      alarms = a->next;
       
       switch (a->action) {
 	
@@ -230,27 +237,30 @@ loopmaster(Alarmloopfunc alarm_looper, Xloopfunc x_looper)
 	break;
 	
        case A_IDLE_SELECT:
-	register_keystrokes((Window)a->data);
+	register_keystrokes((Display *)a->data1, (Window)a->data2);
 	break;
 	
        case A_MOUSE: {
 	 Window root, child;
 	 int root_x, root_y, win_x, win_y;
 	 unsigned mask;
-	 XQueryPointer(display, port.root_window, &root, &child,
-		       &root_x, &root_y, &win_x, &win_y, &mask);
-	 if (root != last_mouse_root
-	     || root_x < last_mouse_x - mouse_sensitivity
-	     || root_x > last_mouse_x + mouse_sensitivity
-	     || root_y < last_mouse_y - mouse_sensitivity
-	     || root_y > last_mouse_y + mouse_sensitivity) {
-	   XEvent event;
-	   event.type = MotionNotify; /* skeletal MotionNotify event */
-	   if (x_looper && last_mouse_root)
-	     ret_val = x_looper(&event, &now);
-	   last_mouse_root = root;
-	   last_mouse_x = root_x;
-	   last_mouse_y = root_y;
+	 int i;
+	 for (i = 0; i < nports; i++) {
+	   XQueryPointer(ports[i].display, ports[i].root_window, &root, &child,
+			 &root_x, &root_y, &win_x, &win_y, &mask);
+	   if (root != ports[i].last_mouse_root
+	       || root_x < ports[i].last_mouse_x - mouse_sensitivity
+	       || root_x > ports[i].last_mouse_x + mouse_sensitivity
+	       || root_y < ports[i].last_mouse_y - mouse_sensitivity
+	       || root_y > ports[i].last_mouse_y + mouse_sensitivity) {
+	     XEvent event;
+	     event.type = MotionNotify; /* skeletal MotionNotify event */
+	     if (x_looper && ports[i].last_mouse_root)
+	       ret_val = x_looper(&event, &now);
+	     ports[i].last_mouse_root = root;
+	     ports[i].last_mouse_x = root_x;
+	     ports[i].last_mouse_y = root_y;
+	   }
 	 }
 	 xwADDTIME(a->timer, a->timer, check_mouse_time);
 	 schedule(a);
@@ -258,7 +268,8 @@ loopmaster(Alarmloopfunc alarm_looper, Xloopfunc x_looper)
        }
        
        default:
-	ret_val = alarm_looper(a, &now);
+	if (alarm_looper)
+	  ret_val = alarm_looper(a, &now);
 	break;
 	
       }
@@ -267,18 +278,18 @@ loopmaster(Alarmloopfunc alarm_looper, Xloopfunc x_looper)
       if (ret_val != 0) return ret_val;
     }
     
-    if (alarms) {
+    if (alarm_sentinel.next != &alarm_sentinel) {
       timeoutptr = &timeout;
-      xwSUBTIME(timeout, alarms->timer, now);
+      xwSUBTIME(timeout, alarm_sentinel.next->timer, now);
     } else
       timeoutptr = 0;
-    
-    pending = XPending(display);
+
+    for (i = pending = 0; !pending && i < nports; i++)
+      pending = XPending(ports[i].display);
     if (!pending) {
       int result;
-      FD_SET(port.x_socket, &xfds);
-      result = select(port.x_socket + 1, &xfds, 0, 0, timeoutptr);
-      pending = (result <= 0 ? 0 : FD_ISSET(port.x_socket, &xfds));
+      xfds = x_socket_set;
+      result = select(max_x_socket + 1, &xfds, 0, 0, timeoutptr);
     }
     
     /* Behave robustly when the system clock is adjusted backwards. The idea:
@@ -290,7 +301,6 @@ loopmaster(Alarmloopfunc alarm_looper, Xloopfunc x_looper)
       xwGETTIME(new_now);
       if (xwTIMEGT(now, new_now)) {
 	xwSUBTIME(new_now, now, new_now);
-	if (timeoutptr) xwSUBTIME(new_now, new_now, timeout);
 	xwSUBTIME(genesis_time, genesis_time, new_now);
 	xwGETTIME(new_now);
       }
@@ -298,10 +308,10 @@ loopmaster(Alarmloopfunc alarm_looper, Xloopfunc x_looper)
     }
     
     /* Handle X events. */
-    if (pending)
-      while (XPending(display)) {
+    for (i = 0; i < nports; i++)
+      while (XPending(ports[i].display)) {
 	XEvent event;
-	XNextEvent(display, &event);
+	XNextEvent(ports[i].display, &event);
 	default_x_processing(&event);
 	if (x_looper) ret_val = x_looper(&event, &now);
 	if (ret_val != 0) return ret_val;
