@@ -13,32 +13,37 @@ pop_up_hand(Hand *h)
   set_slideshow(h, ocurrent->slideshow, 0);
   set_slideshow(h->icon, ocurrent->icon_slideshow, 0);
   h->clock = ocurrent->clock;
-  XMapRaised(display, h->w);
-  XFlush(display);
+  XMapRaised(h->port->display, h->w);
+  XFlush(h->port->display);
 }
 
 
 static int
-switch_options(Options *opt, struct timeval now)
+switch_options(Options *opt, const struct timeval *option_switch_time,
+	       const struct timeval *now)
 {
   Hand *h;
   Alarm *a;
+  int i;
   
   ocurrent = opt;
-  
-  set_all_slideshows(hands, opt->slideshow);
-  set_all_slideshows(icon_hands, opt->icon_slideshow);
+
+  for (i = 0; i < nports; i++) {
+    set_all_slideshows(ports[i].hands, opt->slideshow);
+    set_all_slideshows(ports[i].icon_hands, opt->icon_slideshow);
+  }
   
   if (opt->multiply) {
     a = new_alarm(A_MULTIPLY);
-    xwADDTIME(a->timer, now, opt->multiply_delay);
+    xwADDTIME(a->timer, *now, opt->multiply_delay);
     schedule(a);
-  } else unschedule(A_MULTIPLY);
+  } else
+    unschedule(A_MULTIPLY);
   
   if (opt->clock && !clock_displaying) {
-    draw_all_clocks(&now);
+    draw_all_clocks(now);
     a = new_alarm(A_CLOCK);
-    xwADDTIME(a->timer, now, clock_tick);
+    xwADDTIME(a->timer, *now, clock_tick);
     schedule(a);
     clock_displaying = 1;
   } else if (!opt->clock && clock_displaying) {
@@ -46,20 +51,20 @@ switch_options(Options *opt, struct timeval now)
     erase_all_clocks();
     clock_displaying = 0;
   }
-  
-  for (h = hands; h; h = h->next)
-    if ((opt->never_iconify && !h->mapped) ||
-	(!opt->appear_iconified && h->icon->mapped))
-      XMapRaised(display, h->w);
-  
-  if (opt->beep)
-    XBell(display, 0);
 
-  XFlush(display);
+  for (i = 0; i < nports; i++) {
+    for (h = ports[i].hands; h; h = h->next)
+      if ((opt->never_iconify && !h->mapped) ||
+	  (!opt->appear_iconified && h->icon->mapped))
+	XMapRaised(ports[i].display, h->w);
+    if (opt->beep)
+      XBell(ports[i].display, 0);
+    XFlush(ports[i].display);
+  }
   
   if (opt->next) {
     a = new_alarm(A_NEXT_OPTIONS);
-    xwADDTIME(a->timer, now, opt->next_delay);
+    xwADDTIME(a->timer, *option_switch_time, opt->next_delay);
     schedule(a);
   }
   
@@ -71,19 +76,19 @@ switch_options(Options *opt, struct timeval now)
 
 
 static int
-warn_alarm_loop(Alarm *a, struct timeval *now)
+warn_alarm_loop(Alarm *a, const struct timeval *now)
 {
   switch (a->action) {
     
    case A_MULTIPLY:
     if (active_hands() < ocurrent->max_hands)
-      pop_up_hand(new_hand(NHRandom, NHRandom));
+      pop_up_hand(new_hand(&ports[0], NEW_HAND_RANDOM, NEW_HAND_RANDOM));
     xwADDTIME(a->timer, a->timer, ocurrent->multiply_delay);
     schedule(a);
     break;
     
    case A_NEXT_OPTIONS:
-    return switch_options(ocurrent->next, *now);
+    return switch_options(ocurrent->next, now, now);
     
    case A_IDLE_CHECK:
     return TRAN_REST;
@@ -96,6 +101,7 @@ warn_alarm_loop(Alarm *a, struct timeval *now)
 static int
 check_raise_window(Hand *h)
 {
+  Port *port = h->port;
   Window root, parent, *children = 0;
   unsigned nchildren;
   unsigned i;
@@ -106,11 +112,11 @@ check_raise_window(Hand *h)
   int bad_overlap;
   
   /* find our geometry */
-  XGetGeometry(display, h->root_child, &root, &hx, &hy,
+  XGetGeometry(port->display, h->root_child, &root, &hx, &hy,
 	       (unsigned *)&hwidth, (unsigned *)&hheight, &border, &depth);
   
   /* find our position in the stacking order */
-  if (!XQueryTree(display, port.root_window, &root, &parent,
+  if (!XQueryTree(port->display, port->root_window, &root, &parent,
 		  &children, &nchildren)) {
     if (children) XFree(children);
     return 0;
@@ -123,11 +129,11 @@ check_raise_window(Hand *h)
   bad_overlap = 0;
   for (i++; i < nchildren && !bad_overlap; i++) {
     /* overlap by other hands is OK */
-    for (trav = hands; trav; trav = trav->next)
+    for (trav = port->hands; trav; trav = trav->next)
       if (children[i] == trav->root_child)
 	goto overlap_ok;
     /* check to see that this window actually overlaps us */
-    XGetGeometry(display, children[i], &root, &ox, &oy,
+    XGetGeometry(port->display, children[i], &root, &ox, &oy,
 		 (unsigned *)&owidth, (unsigned *)&oheight, &border, &depth);
     if (ox > hx + hwidth || oy > hy + hheight
 	|| ox + owidth < hx || oy + oheight < hy)
@@ -142,17 +148,19 @@ check_raise_window(Hand *h)
 }
 
 static int
-warn_x_loop(XEvent *e, struct timeval *now)
+warn_x_loop(XEvent *e, const struct timeval *now)
 {
   Alarm *a;
   Hand *h;
   switch (e->type) {
     
-   case MapNotify:
-    h = window_to_hand(e->xunmap.window, 1);
-    if (h && h->is_icon && ocurrent->never_iconify)
-      XMapRaised(display, h->icon->w);
-    break;
+   case MapNotify: {
+     Port *port = find_port(e->xunmap.display);
+     h = window_to_hand(port, e->xunmap.window, 1);
+     if (h && h->is_icon && ocurrent->never_iconify)
+       XMapRaised(port->display, h->icon->w);
+     break;
+   }
     
    case ClientMessage:
     /* Check for window manager deleting last xwrits window */
@@ -162,53 +170,77 @@ warn_x_loop(XEvent *e, struct timeval *now)
     
    case ButtonPress:
     /* OK; we can rest now. */
+    /* 10.Aug.1999 - treat the mouse click as a keypress */
+    last_key_time = *now;
     return TRAN_REST;
     
    case KeyPress:
+   case MotionNotify:
+    last_key_time = *now;
     if ((a = grab_alarm(A_IDLE_CHECK))) {
-      last_key_time = *now;
-      xwADDTIME(a->timer, last_key_time, idle_time);
+      xwADDTIME(a->timer, last_key_time, warn_idle_time);
       schedule(a);
     }
     break;
     
-   case VisibilityNotify:
-    if ((h = window_to_hand(e->xvisibility.window, 0)) && h->obscured
-	&& ocurrent->top)
-      if (check_raise_window(h))
-	XRaiseWindow(display, h->w);
-    break;
-    
+   case VisibilityNotify: {
+     Port *port = find_port(e->xvisibility.display);
+     h = window_to_hand(port, e->xvisibility.window, 0);
+     if (h && h->obscured && ocurrent->top)
+       if (check_raise_window(h))
+	 XRaiseWindow(port->display, h->w);
+     break;
+   }
+   
   }
   return 0;
 }
 
 
 int
-warn(int was_lock)
+warn(int was_lock, Options *onormal)
 {
-  struct timeval now;
-  int val;
+  struct timeval now, option_switch_time;
+  int i, val;
   
   clock_displaying = 0;
   clock_zero_time = first_warn_time;
-  
+
+  /* find correct options based on elapsed time since first_warn_time */
   xwGETTIME(now);
-  val = switch_options(ocurrent, now);
+  ocurrent = onormal;
+  option_switch_time = first_warn_time;
+  while (ocurrent->next) {
+    struct timeval next;
+    xwADDTIME(next, option_switch_time, ocurrent->next_delay);
+    if (xwTIMEGT(next, now))
+      break;
+    option_switch_time = next;
+    ocurrent = ocurrent->next;
+  }
+
+  /* switch to those options */
+  /* This will always set_slideshows: good -- later set_slideshows will start
+     from scratch. */
+  val = switch_options(ocurrent, &option_switch_time, &now);
+
+  /* lock now if we should */
   if (val == TRAN_LOCK && !was_lock)
-    return TRAN_LOCK;
+    goto done;
+
+  for (i = 0; i < nports; i++)
+    pop_up_hand(ports[i].hands);
   
-  pop_up_hand(hands);
-  
-  if (check_idle && !xwTIMELEQ0(idle_time)) {
+  if (check_idle) {
     Alarm *a = new_alarm(A_IDLE_CHECK);
-    xwADDTIME(a->timer, last_key_time, idle_time);
+    xwADDTIME(a->timer, last_key_time, warn_idle_time);
     schedule(a);
   }
   
   val = loopmaster(warn_alarm_loop, warn_x_loop);
+
+ done:
   unschedule(A_FLASH | A_MULTIPLY | A_CLOCK | A_NEXT_OPTIONS | A_IDLE_CHECK);
-  
   erase_all_clocks();
   return val;
 }
